package org.cruk.cri.caarray;

import edu.georgetown.pir.Organism;
import gov.nih.nci.caarray.application.file.InvalidFileException;
import gov.nih.nci.caarray.application.project.InconsistentProjectStateException;
import gov.nih.nci.caarray.application.project.ProposalWorkflowException;
import gov.nih.nci.caarray.domain.array.ArrayDesign;
import gov.nih.nci.caarray.domain.contact.Organization;
import gov.nih.nci.caarray.domain.file.CaArrayFile;
import gov.nih.nci.caarray.domain.file.FileType;
import gov.nih.nci.caarray.domain.project.AssayType;
import gov.nih.nci.caarray.domain.project.Project;
import gov.nih.nci.caarray.domain.vocabulary.Term;
import gov.nih.nci.caarray.domain.vocabulary.TermSource;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;
import java.util.Set;

/**
 * Service facade providing remote access to a number of retrieval and update
 * methods not available through the remotely available EJBs provided by
 * caArray.
 *
 * @author eldrid01
 */
public interface CaArrayServiceFacade
{
    /**
     * The JNDI name to look up the remote <code>CaArrayServiceFacade</code> service.
     */
    String JNDI_NAME = "caarray/CaArrayServiceFacadeBean/remote";

    /**
     * @return the list of all providers in the system.
     */
    List<Organization> getAllProviders();

    /**
     * Returns the list of ArrayDesigns with the given provider and assay type.
     *
     * @param provider the provider
     * @param assayType the assay type
     * @return the List&lt;ArrayDesign&gt; of the array designs whose provider is the given provider
     */
    List<ArrayDesign> getImportedArrayDesigns(Organization provider, AssayType assayType);

    /**
     * Returns the requested term source, if it exists.
     *
     * @param name name of the source
     * @param version version to retrieve. If null, then will retrieve the term source with given name and no version
     * @return the matching source, or null if no matching source exists.
     */
    TermSource getSource(String name, String version);

    /**
     * Returns the term with the given value from the given term source if it exists, or null if it does not.
     *
     * @param source the source to which the term should belong
     * @param value value of the term (case insensitive)
     * @return the matching term or null if no term with that value exists
     */
    Term getTerm(TermSource source, String value);

    /**
     * Returns the organism with the given id.
     *
     * @param id the id of the desired organism
     * @return the organism with given id or null if none found.
     */
    Organism getOrganism(Long id);

    /**
     * Returns all Organisms.
     *
     * @return the List&lt;Organism&gt; in the system
     */
    List<Organism> getOrganisms();

    /**
     * Creates a new project with the given title, description and array design.
     *
     * @param title
     * @param description
     * @param arrayDesignId
     * @return
     * @throws ProposalWorkflowException
     * @throws InconsistentProjectStateException
     */
    Project createProject(String title, String description, long arrayDesignId) throws ProposalWorkflowException, InconsistentProjectStateException;

    /**
     * Deletes a project. The project must be in the draft state.
     *
     * @param project the project to save
     * @throws ProposalWorkflowException if the project cannot currently be deleted because it is not a draft
     */
    void deleteProject(Project project) throws ProposalWorkflowException;

    /**
     * Sets the experiment title for the given project.
     *
     * @param project the project.
     * @param title the title of the experiment.
     */
    void setExperimentTitle(Project project, String title) throws ProposalWorkflowException, InconsistentProjectStateException;

    /**
     * Uploads a file to the project.
     *
     * @param project
     * @param file
     * @param isSupplemental
     * @throws ProposalWorkflowException
     * @throws IOException
     * @throws InconsistentProjectStateException
     * @throws InvalidFileException
     */
    void uploadFile(Project project, File file, boolean isSupplemental)
        throws ProposalWorkflowException, IOException, InconsistentProjectStateException, InvalidFileException;

    /**
     * Marks the given uploaded file as supplemental.
     *
     * @param project the project.
     * @param filename the name of the file to be marked as supplemental.
     * @throws FileNotFoundException
     */
    void addSupplementalFile(Project project, String filename) throws FileNotFoundException;

    /**
     * Change the file type of the given CaArrayFile object.
     *
     * @param caArrayFile the file.
     * @param fileType the new type.
     */
    void changeFileType(CaArrayFile caArrayFile, FileType fileType);

    /**
     * Removes a file from caArray file storage.
     *
     * @param file the caArrayFile to remove
     */
    void removeFile(CaArrayFile caArrayFile);

    /**
     * Initiates the validation of a set of files within the given project.
     *
     * @param project
     * @param files
     */
    void validateFiles(Project project, Set<CaArrayFile> files);

    /**
     * Initiates the import of a set of files within the given project.
     *
     * @param project
     * @param files
     */
    void importFiles(Project project, Set<CaArrayFile> files);
}
